﻿Option Strict On
Option Infer On

Imports Helpers.Mathematics
Imports Helpers.Direct2D
Imports SlimDX.Direct2D
Imports SlimDX.DirectWrite
Imports System.IO
Imports System.Reflection

Friend MustInherit Class Tower
    Inherits Entity

    Private target As Creep
    Private m_radius As Integer
    Private m_damage As Integer
    Private m_location As PointF
    Private m_shotSpeed As Integer
    Private m_speedCounter As Integer
    Private shotSound As SoundEffect
    Private m_level As Integer
    Private m_cost As Integer
    Private m_build As Boolean
    Private sSound As String
    Private m_Image As Bitmap
    Private tColor As Color
    Private armLength As Single
    Private armWidth As Single

    Private Shared ini As INIFile

    Friend Shared Sub Init()
        ini = New INIFile(Path.Combine(Game.getEnvironmentPath("settings"), "towers.ini"))
    End Sub

    Protected Shared Function getValue(key As String) As String
        Return ini.getValue(key)
    End Function

    Friend Shared Function GetAllTowers() As Type()
        Dim currentAssembly = Assembly.GetExecutingAssembly
        Dim currentTypes() = currentAssembly.GetTypes
        Dim towers As New List(Of Type)
        For Each t In currentTypes
            Dim constructors() = t.GetConstructors
            Dim hasRightConstructor = False
            For Each c In constructors
                If c.GetParameters.Count = 0 Then
                    hasRightConstructor = True
                    Exit For
                End If
            Next
            If hasRightConstructor AndAlso t.IsSubclassOf(GetType(Tower)) Then towers.Add(t)
        Next
        Return towers.ToArray
    End Function

    Friend ReadOnly Property radius As Integer
        Get
            Return m_radius + Convert.ToInt32(m_radius * 0.07 * (m_level - 1))
        End Get
    End Property

    Friend ReadOnly Property damage As Integer
        Get
            Return m_damage + Convert.ToInt32(m_damage * 0.4 * (m_level - 1))
        End Get
    End Property

    Protected ReadOnly Property shotSpeed As Integer
        Get
            Return Math.Max(m_shotSpeed - Convert.ToInt32(m_shotSpeed * 0.04 * (m_level - 1)), m_shotSpeed \ 5)
        End Get
    End Property

    Protected ReadOnly Property angle As Single
        Get
            If target Is Nothing Then Return Math.PI / 2.0
            Return Convert.ToSingle(MathHelper.getAngle(New Vector2(m_location), New Vector2(target.getLocation)))
        End Get
    End Property

    Friend ReadOnly Property level As Integer
        Get
            Return m_level
        End Get
    End Property

    Friend MustOverride ReadOnly Property name As String

    Friend Shared ReadOnly Property maxLevel As Integer
        Get
            Return 10
        End Get
    End Property

    Friend Shared ReadOnly Property size As SizeF
        Get
            Return New SizeF(24, 24)
        End Get
    End Property

    Friend Overrides Function getSize() As SizeF
        Return Tower.size
    End Function

    Friend Overrides Function getLocation() As PointF
        Return m_location
    End Function

    Friend Function getTarget() As Creep
        Return target
    End Function

    Friend Function trySetTarget(c As Creep) As Boolean
        If MathHelper.getDistance(New Vector2(getLocation), New Vector2(c.getLocation)) <= radius Then
            target = c
            Return True
        Else
            Return False
        End If
    End Function

    Friend Overrides Sub doStep()
        If Not world.getEntities.Contains(target) Then target = Nothing
        If target IsNot Nothing AndAlso MathHelper.getDistance(New Vector2(getLocation), New Vector2(target.getLocation)) > radius Then target = Nothing
        If target Is Nothing Then
            Dim entities() = world.getEntities
            Dim distance = Double.MaxValue
            Dim index = -1
            For i = 0 To entities.Length - 1
                Dim e = entities(i)
                If TypeOf e Is Creep Then
                    Dim c = DirectCast(e, Creep)
                    Dim cDist = MathHelper.getDistance(New Vector2(c.getLocation), New Vector2(m_location))
                    If cDist < radius AndAlso cDist < distance Then
                        distance = cDist
                        index = i
                    End If
                End If
            Next
            If index <> -1 Then target = DirectCast(entities(index), Creep)
        End If
        If m_speedCounter < shotSpeed Then m_speedCounter += 1
        If m_speedCounter >= shotSpeed AndAlso target IsNot Nothing Then
            m_speedCounter = 0
            shotSound.play()
            shoot()
        End If
    End Sub

    Friend Overrides Sub onDespawned()
        MyBase.onDespawned()
        shotSound.unload()
    End Sub

    Protected MustOverride Sub shoot()

    Friend Function getLevelCost() As Integer
        If m_level >= maxLevel Then Return 0
        Return Convert.ToInt32(m_cost * (m_level + 1) * 0.2)
    End Function

    Friend Function levelUp() As Boolean
        If Not m_build Then Return False
        If m_level >= maxLevel Then Return False
        Dim money = getLevelCost()
        If world.pay(money) Then
            m_level += 1
            Return True
        Else
            Return False
        End If
    End Function

    Friend Function build(w As World, loc As PointF) As Boolean
        If m_build Then Return True
        Dim rect As New RectangleF(loc.X - size.Width / 2.0F, loc.Y - size.Height / 2.0F, size.Width, size.Height)
        If Not w.canBuild(rect) Then Return False
        If w.pay(m_cost) Then
            w.spawnEntity(Me)
            m_build = True
            shotSound = New SoundEffect(sSound)
            m_location = loc
            Return True
        Else
            Return False
        End If
    End Function

    Sub New(radius As Integer, damage As Integer, speed As Integer, sSound As String, cost As Integer, image As Bitmap, tColor As Color, aLength As Single, aWidth As Single)
        m_radius = radius
        m_damage = damage
        m_shotSpeed = speed
        m_speedCounter = speed
        m_cost = cost
        m_build = False
        Me.sSound = sSound
        m_Image = image
        Me.tColor = tColor
        armLength = aLength
        armWidth = aWidth
        m_level = 1
    End Sub

    Friend Overrides Sub draw(e As D2DControlset.D2DDrawEventArgs, offset As PointF)
        Dim h As New Direct2DHelper(e.Factory, e.Target, e.TextFactory)
        Dim brushName = "TowerColor" & tColor.ToString
        If Not ColorManager.BrushRegistered(brushName) Then ColorManager.RegisterBrush(brushName, New SolidColorBrush(e.Target, New SlimDX.Color4(tColor)))
        Dim vm As New Vector2(getLocation.X + offset.X, getLocation.Y + offset.Y)
        h.DrawCircle(vm, size.Width / 2.0F, ColorManager.GetBrush(brushName), 3)
        h.FillCircle(vm, armWidth / 2.0F, ColorManager.GetBrush(brushName))
        Dim v2 As Vector2 = MathHelper.getPointOnCircle(vm, angle, armLength)
        e.Target.DrawLine(ColorManager.GetBrush(brushName), vm.toPointF, v2.toPointF, armWidth)
        Dim x1 = getLocation.X + offset.X - 8
        Dim x2 = getLocation.X + offset.X + 8
        Dim yStart = getLocation.Y + offset.Y - size.Height / 2.0F - 8
        If level < 5 Then
            For i = 1 To level
                Dim p As New PointF(x1, yStart - i * 5)
                drawArrow(e, p)
            Next
        Else
            h.FillStar(New PointF(x1, yStart - 10), 8, 90, ColorManager.GetBrush("defaultWhite"), 5)
            If level < 10 Then
                For i = 1 To level - 5
                    Dim p As New PointF(x2, yStart - i * 5)
                    drawArrow(e, p)
                Next
            Else
                h.FillStar(New PointF(x2, yStart - 10), 8, 90, ColorManager.GetBrush("defaultWhite"), 5)
            End If
        End If
    End Sub

    Private Sub drawArrow(e As D2DControlset.D2DDrawEventArgs, p As PointF)
        Dim f = Game.getSizeFactor
        Dim pg As New PathGeometry(e.Factory)
        Dim sink = pg.Open
        sink.BeginFigure(New PointF(p.X - 6, p.Y + 3), FigureBegin.Hollow)
        sink.AddLine(p)
        sink.AddLine(New PointF(p.X + 6, p.Y + 3))
        sink.EndFigure(FigureEnd.Open)
        sink.Close()
        e.Target.DrawGeometry(pg, ColorManager.GetBrush("defaultWhite"), 2)
        sink.Dispose()
        pg.Dispose()
    End Sub

    Friend Function getImage() As Bitmap
        Return m_Image
    End Function
End Class

Friend MustInherit Class ProjectileTower
    Inherits Tower

    Private projectileColor As Color
    Private projectileMove As Double
    Private projectileSize As SizeF
    Private projectileHoming As Boolean
    Private projectileRetarget As Boolean

    Protected Overrides Sub shoot()
        Dim p As Projectile
        If projectileHoming Then
            p = New HomingProjectile(projectileMove, projectileColor, getLocation, projectileSize, damage, getTarget, projectileRetarget, Me.GetType)
        Else
            Dim v As Vector2 = MathHelper.getPointOnCircle(New Vector2(), angle, projectileMove)
            p = New Projectile(v.toPointF, projectileColor, getLocation, projectileSize, damage, True, Me.GetType)
        End If
        p.world = world
        world.spawnEntity(p)
    End Sub

    Sub New(radius As Integer, damage As Integer, speed As Integer, sSound As String, cost As Integer, pColor As Color, pMove As Double, pSize As SizeF, pHoming As Boolean, image As Bitmap, tColor As Color, aLength As Single, aWidth As Single, retarget As Boolean)
        MyBase.New(radius, damage, speed, sSound, cost, image, tColor, aLength, aWidth)
        projectileColor = pColor
        projectileMove = pMove
        projectileSize = pSize
        projectileHoming = pHoming
        projectileRetarget = retarget
    End Sub
End Class

Friend Class GunnerTower
    Inherits ProjectileTower

    Friend Shared Function getName() As String
        Return Tower.getValue("tower1")
    End Function

    Friend Shared Function getCosts() As Integer
        Return Integer.Parse(Tower.getValue("t01_cost"))
    End Function

    Friend Shared Function getRadius() As Integer
        Return Integer.Parse(Tower.getValue("t01_radius"))
    End Function

    Friend Shared Function getDamage() As Integer
        Return Integer.Parse(Tower.getValue("t01_damage"))
    End Function

    Friend Shared Function getSpeed() As Integer
        Return Integer.Parse(Tower.getValue("t01_shotspeed"))
    End Function

    Sub New()
        MyBase.New(getRadius, getDamage, getSpeed, Tower.getValue("t01_sound"), getCosts, GetBaseColor, 10, New SizeF(8, 4), False, Game.towerImages(0), GetBaseColor, size.Width * 0.85F, 4, False)
    End Sub

    Public Shared Function GetBaseColor() As Color
        Return Color.Lime
    End Function

    Friend Overrides ReadOnly Property name As String
        Get
            Return getName().Replace(Environment.NewLine, String.Empty)
        End Get
    End Property
End Class

Friend Class RocketTower
    Inherits ProjectileTower

    Friend Shared Function getName() As String
        Return Tower.getValue("tower2")
    End Function

    Friend Shared Function getCosts() As Integer
        Return Integer.Parse(Tower.getValue("t02_cost"))
    End Function

    Friend Shared Function getRadius() As Integer
        Return Integer.Parse(Tower.getValue("t02_radius"))
    End Function

    Friend Shared Function getDamage() As Integer
        Return Integer.Parse(Tower.getValue("t02_damage"))
    End Function

    Friend Shared Function getSpeed() As Integer
        Return Integer.Parse(Tower.getValue("t02_shotspeed"))
    End Function

    Sub New()
        MyBase.New(getRadius, getDamage, getSpeed, Tower.getValue("t02_sound"), getCosts, GetBaseColor, 3, New Size(12, 7), True, Game.towerImages(1), GetBaseColor, size.Width * 0.7F, 10, True)
    End Sub

    Public Shared Function GetBaseColor() As Color
        Return Color.DarkRed
    End Function

    Friend Overrides ReadOnly Property name As String
        Get
            Return getName().Replace(Environment.NewLine, String.Empty)
        End Get
    End Property
End Class

Friend Class SniperTower
    Inherits ProjectileTower

    Friend Shared Function getName() As String
        Return Tower.getValue("tower3")
    End Function

    Friend Shared Function getCosts() As Integer
        Return Integer.Parse(Tower.getValue("t03_cost"))
    End Function

    Friend Shared Function getRadius() As Integer
        Return Integer.Parse(Tower.getValue("t03_radius"))
    End Function

    Friend Shared Function getDamage() As Integer
        Return Integer.Parse(Tower.getValue("t03_damage"))
    End Function

    Friend Shared Function getSpeed() As Integer
        Return Integer.Parse(Tower.getValue("t03_shotspeed"))
    End Function

    Sub New()
        MyBase.New(getRadius, getDamage, getSpeed, Tower.getValue("t03_sound"), getCosts, GetBaseColor, 25, New Size(12, 2), True, Game.towerImages(2), GetBaseColor, size.Width * 0.9F, 3, False)
    End Sub

    Public Shared Function GetBaseColor() As Color
        Return Color.Purple
    End Function

    Friend Overrides ReadOnly Property name As String
        Get
            Return getName().Replace(Environment.NewLine, String.Empty)
        End Get
    End Property
End Class

Friend MustInherit Class LaserTower
    Inherits Tower

    Private laserDuration As Integer
    Private laserCounter As Integer
    Private shooting As Boolean
    Private shootTimes() As Integer
    Private laserWidth As Single
    Private laserRadius As Single
    Private laserColor As Color

    Friend Overrides Sub doStep()
        MyBase.doStep()
        If laserCounter < laserDuration Then laserCounter += 1
        If shooting AndAlso isShootStep() AndAlso (getTarget() IsNot Nothing) Then
            Dim ex As New Explosion(getTarget.getLocation, {laserColor}, laserColor, 15, 0, False, False)
            world.spawnEntity(ex)
            getTarget.takeDamage(1, Me.GetType)
        End If
        If laserCounter = laserDuration Then shooting = False
    End Sub

    Private Function isShootStep() As Boolean
        For i = 0 To shootTimes.Length - 1
            If laserCounter = shootTimes(i) Then Return True
        Next
        Return False
    End Function

    Protected Overrides Sub shoot()
        laserCounter = 0
        shooting = True
    End Sub

    Friend Function laserVisible() As Boolean
        Return laserCounter <= laserDuration
    End Function

    Sub New(radius As Integer, damage As Integer, speed As Integer, sSound As String, cost As Integer, lDuration As Integer, image As Bitmap, tColor As Color, aLength As Single, aWidth As Single, lWidth As Single, lRadius As Single, lColor As Color)
        MyBase.New(radius, damage, speed, sSound, cost, image, tColor, aLength, aWidth)
        laserDuration = lDuration
        laserWidth = lWidth
        laserRadius = lRadius
        laserColor = lColor
        setShootTimes()
    End Sub

    Private Sub setShootTimes()
        shootTimes = New Integer(damage - 1) {}
        Dim timeStep As Double = laserDuration / damage
        For i = 1 To damage - 1
            shootTimes(i - 1) = Convert.ToInt32(timeStep * i)
        Next
        shootTimes(damage - 1) = laserDuration
    End Sub

    Friend Overrides Sub draw(e As D2DControlset.D2DDrawEventArgs, offset As PointF)
        MyBase.draw(e, offset)
        If Not shooting Then Exit Sub
        Dim vm As New Vector2(getLocation.X + offset.X, getLocation.Y + offset.Y)
        Dim vl As Vector2 = MathHelper.getPointOnCircle(vm, angle, laserRadius)
        Dim b As New SolidColorBrush(e.Target, New SlimDX.Color4(laserColor))
        e.Target.DrawLine(b, vm.toPointF, vl.toPointF, laserWidth)
        b.Dispose()
    End Sub
End Class

Friend Class StreamTower
    Inherits LaserTower

    Friend Shared Function getName() As String
        Return Tower.getValue("tower4")
    End Function

    Friend Shared Function getCosts() As Integer
        Return Integer.Parse(Tower.getValue("t04_cost"))
    End Function

    Friend Shared Function getRadius() As Integer
        Return Integer.Parse(Tower.getValue("t04_radius"))
    End Function

    Friend Shared Function getDamage() As Integer
        Return Integer.Parse(Tower.getValue("t04_damage"))
    End Function

    Friend Shared Function getSpeed() As Integer
        Return Integer.Parse(Tower.getValue("t04_shotspeed"))
    End Function

    Sub New()
        MyBase.New(getRadius, getDamage, getSpeed, Tower.getValue("t04_sound"), getCosts, 40, Game.towerImages(3), GetBaseColor, size.Width * 0.8F, 4, 3, 300, GetBaseColor)
    End Sub

    Public Shared Function GetBaseColor() As Color
        Return Color.Blue
    End Function

    Friend Overrides ReadOnly Property name As String
        Get
            Return getName().Replace(Environment.NewLine, String.Empty)
        End Get
    End Property
End Class

Friend Class FlashTower
    Inherits LaserTower

    Friend Shared Function getName() As String
        Return Tower.getValue("tower5")
    End Function

    Friend Shared Function getCosts() As Integer
        Return Integer.Parse(Tower.getValue("t05_cost"))
    End Function

    Friend Shared Function getRadius() As Integer
        Return Integer.Parse(Tower.getValue("t05_radius"))
    End Function

    Friend Shared Function getDamage() As Integer
        Return Integer.Parse(Tower.getValue("t05_damage"))
    End Function

    Friend Shared Function getSpeed() As Integer
        Return Integer.Parse(Tower.getValue("t05_shotspeed"))
    End Function

    Sub New()
        MyBase.New(getRadius, getDamage, getSpeed, Tower.getValue("t05_sound"), getCosts, 10, Game.towerImages(4), GetBaseColor, size.Width * 0.75F, 6, 5, 160, GetBaseColor)
    End Sub

    Public Shared Function GetBaseColor() As Color
        Return Color.Yellow
    End Function

    Friend Overrides ReadOnly Property name As String
        Get
            Return getName().Replace(Environment.NewLine, String.Empty)
        End Get
    End Property
End Class

Friend Class AoETower : Inherits Tower

    Friend Shared Function getName() As String
        Return Tower.getValue("tower6")
    End Function

    Friend Shared Function getCosts() As Integer
        Return Integer.Parse(Tower.getValue("t06_cost"))
    End Function

    Friend Shared Function getRadius() As Integer
        Return Integer.Parse(Tower.getValue("t06_radius"))
    End Function

    Friend Shared Function getDamage() As Integer
        Return Integer.Parse(Tower.getValue("t06_damage"))
    End Function

    Friend Shared Function getSpeed() As Integer
        Return Integer.Parse(Tower.getValue("t06_shotspeed"))
    End Function

    Sub New()
        MyBase.New(getRadius, getDamage, getSpeed, Tower.getValue("t06_sound"), getCosts, Game.towerImages(5), GetBaseColor, 0, 0)
    End Sub

    Public Shared Function GetBaseColor() As Color
        Return Color.Green
    End Function

    Friend Overrides ReadOnly Property name As String
        Get
            Return getName().Replace(Environment.NewLine, String.Empty)
        End Get
    End Property

    Protected Overrides Sub shoot()
        For i = 0 To 7
            Dim a = i * Math.PI / 4
            Dim v = MathHelper.getPointOnCircle(New Helpers.Mathematics.Vector2(0, 0), a, 2)
            Dim bomb As New BombProjectile(v.toPointF, GetBaseColor, getLocation, New SizeF(10, 10), damage, radius \ 4, radius / 2, GetType(AoETower))
            world.spawnEntity(bomb)
        Next
    End Sub

    Friend Overrides Sub draw(e As D2DControlset.D2DDrawEventArgs, offset As PointF)
        MyBase.draw(e, offset)
        Dim h As New Direct2DHelper(e.Factory, e.Target, e.TextFactory)
        Dim m As New PointF(getLocation.X + offset.X, getLocation.Y + offset.Y)
        Dim brushName = "TowerColor" & GetBaseColor.ToString
        h.FillStar(m, 8, 90, ColorManager.GetBrush(brushName), 6)
    End Sub
End Class

Friend Class SlownessTower : Inherits Tower
    Private showField As Boolean
    Private effectTime As Integer

    Friend Shared Function getName() As String
        Return Tower.getValue("tower7")
    End Function

    Friend Shared Function getCosts() As Integer
        Return Integer.Parse(Tower.getValue("t07_cost"))
    End Function

    Friend Shared Function getRadius() As Integer
        Return Integer.Parse(Tower.getValue("t07_radius"))
    End Function

    Friend Shared Function getDamage() As Integer
        Return Integer.Parse(Tower.getValue("t07_damage"))
    End Function

    Friend Shared Function getSpeed() As Integer
        Return Integer.Parse(Tower.getValue("t07_shotspeed"))
    End Function

    Sub New()
        MyBase.New(getRadius, getDamage, getSpeed, Tower.getValue("t07_sound"), getCosts, Game.towerImages(5), Color.White, 0, 0)
    End Sub

    Friend Overrides ReadOnly Property name As String
        Get
            Return getName().Replace(Environment.NewLine, String.Empty)
        End Get
    End Property

    Protected Overrides Sub shoot()
        Dim ents() = world.getEntities
        For Each ent In ents
            If TypeOf ent Is Creep Then
                ent.InUse = True
                Dim c = DirectCast(ent, Creep)
                If MathHelper.getDistance(New Helpers.Mathematics.Vector2(c.getLocation), New Helpers.Mathematics.Vector2(getLocation)) <= radius Then c.SlowDown(0.25, 120)
                ent.InUse = False
            End If
        Next
        showField = True
        effectTime = 0
    End Sub

    Friend Overrides Sub draw(e As D2DControlset.D2DDrawEventArgs, offset As PointF)
        MyBase.draw(e, offset)
        Dim h As New Direct2DHelper(e.Factory, e.Target, e.TextFactory)
        Dim vm As New Vector2(getLocation.X + offset.X, getLocation.Y + offset.Y)
        Dim brushName = "TowerColor" & Color.White.ToString
        h.FillCircle(vm, 7, ColorManager.GetBrush(brushName))
        If showField Then
            Dim s1 As New GradientStop
            s1.Color = New SlimDX.Color4(0.05F, 1, 1, 1)
            s1.Position = 0
            Dim s2 As New GradientStop
            s2.Color = New SlimDX.Color4(0.07F, 1, 1, 1)
            s2.Position = 0.7F
            Dim s3 As New GradientStop
            s3.Color = New SlimDX.Color4(0.15F, 1, 1, 1)
            s3.Position = 0.9F
            Dim s4 As New GradientStop
            s4.Color = New SlimDX.Color4(0, 1, 1, 1)
            s4.Position = 1
            Dim stops As New GradientStopCollection(e.Target, {s1, s2, s3, s4}, Gamma.Linear, ExtendMode.Clamp)
            Dim props As New RadialGradientBrushProperties
            props.CenterPoint = New PointF(getLocation.X + offset.X, getLocation.Y + offset.Y)
            Dim r = radius / 60.0F * effectTime
            props.HorizontalRadius = r
            props.VerticalRadius = r
            Dim b As New RadialGradientBrush(e.Target, stops, props)
            e.Target.FillEllipse(b, New Ellipse() With {.Center = props.CenterPoint, .RadiusX = r, .RadiusY = r})
            b.Dispose()
            stops.Dispose()
            effectTime += 1
            If effectTime > 60 Then showField = False
        End If
    End Sub
End Class

Friend Class Projectile
    Inherits Particle

    Protected damage As Integer
    Private lastPoints As Queue(Of PointF)
    Private canHit As Boolean
    Protected towerType As Type

    Friend Overrides Sub doStep()
        MyBase.doStep()
        lastPoints.Enqueue(getLocation)
        If lastPoints.Count > Convert.ToInt32(30 / MathHelper.getDistance(New Vector2(getLocation), New Vector2(getLocation.X + moveVec.X, getLocation.Y + moveVec.Y))) Then lastPoints.Dequeue()
        If world Is Nothing Then Return
        If canHit Then
            For Each e In world.getEntities
                If TypeOf e Is Creep Then
                    e.InUse = True
                    Dim c As Creep = DirectCast(e, Creep)
                    If c.containsPoint(m_Location) Then
                        Dim ex As New Explosion(m_Location, {color}, color, 20 + damage, 0.0F, False, False)
                        world.spawnEntity(ex)
                        world.despawnEntity(Me)
                        c.takeDamage(damage, towerType)
                        e.InUse = False
                        Exit For
                    End If
                    e.InUse = False
                End If
            Next
        End If
    End Sub

    Sub New(v As PointF, c As Color, loc As PointF, s As SizeF, d As Integer, hits As Boolean, tower As Type)
        MyBase.New(v, c, loc, s, False)
        damage = d
        lastPoints = New Queue(Of PointF)
        canHit = hits
        towerType = tower
    End Sub

    Friend Overrides Sub draw(e As D2DControlset.D2DDrawEventArgs, offset As PointF)
        Dim h As New Direct2DHelper(e.Factory, e.Target, e.TextFactory)
        Dim brushName = "ProjectileColor" & color.ToString
        If Not ColorManager.BrushRegistered(brushName) Then ColorManager.RegisterBrush(brushName, New SolidColorBrush(e.Target, New SlimDX.Color4(color)))
        ColorManager.GetBrush(brushName).Opacity = 1
        Dim mv As New Vector2(getLocation.X + offset.X, getLocation.Y + offset.Y)
        Dim el As New Ellipse
        el.Center = mv.toPointF
        el.RadiusX = getSize.Width / 2.0F
        el.RadiusY = getSize.Height / 2.0F
        Dim a As Double = MathHelper.getAngle(mv, New Vector2(mv.X + moveVec.X, mv.Y + moveVec.Y))
        h.FillEllipse(el, Convert.ToSingle(MathHelper.getDeg(a)), ColorManager.GetBrush(brushName))
        Dim tempPoints As New List(Of PointF)
        For i = 0 To lastPoints.Count - 1
            If i >= lastPoints.Count Then Exit For
            Try
                tempPoints.Add(lastPoints.ElementAt(i))
            Catch : End Try
        Next
        Dim points() As PointF = tempPoints.ToArray
        Array.Reverse(points)
        For i = 0 To points.Length - 1
            points(i).X += offset.X
            points(i).Y += offset.Y
        Next
        ColorManager.GetBrush(brushName).Opacity = 0.6F
        For i = 0 To points.Length - 2
            Dim w As Single = (points.Length - i) * (getSize.Height * 1.2F / points.Length)
            e.Target.DrawLine(ColorManager.GetBrush(brushName), points(i), points(i + 1), w)
            ColorManager.GetBrush(brushName).Opacity -= 0.005F
        Next
    End Sub
End Class

Friend Class HomingProjectile
    Inherits Projectile

    Private target As Creep
    Private moveSpeed As Double
    Private retarget As Boolean

    Friend Overrides Sub doStep()
        If Not world.getEntities.Contains(target) Then target = Nothing
        If target Is Nothing Then
            Dim index As Integer = -1
            Dim entities() As Entity = world.getEntities
            If retarget Then
                Dim distance As Double = Double.MaxValue
                For i = 0 To entities.Length - 1
                    Dim e = entities(i)
                    If TypeOf e Is Creep Then
                        e.InUse = True
                        Dim c As Creep = DirectCast(e, Creep)
                        Dim cDist As Double = MathHelper.getDistance(New Vector2(m_Location), New Vector2(c.getLocation))
                        If cDist < distance Then
                            distance = cDist
                            index = i
                        End If
                        e.InUse = False
                    End If
                Next
            End If
            If index <> -1 Then
                target = DirectCast(entities(index), Creep)
            Else
                Dim ex As New Explosion(m_Location, {color}, color, 20 + damage, 0.0F, False, False)
                world.spawnEntity(ex)
                world.despawnEntity(Me)
                Exit Sub
            End If
        End If
        Dim a As Double = getNewAngle(MathHelper.getRad(5))
        Dim v As Vector2 = MathHelper.getPointOnCircle(New Vector2(), a, moveSpeed)
        moveVec = v.toPointF
        MyBase.doStep()
    End Sub

    Private Function getNewAngle(max As Double) As Double
        Dim oldA = MathHelper.getAngle(New Vector2(0, 0), New Vector2(moveVec))
        Dim newA = MathHelper.getAngle(New Vector2(m_Location), New Vector2(target.getLocation))
        Dim round = 2 * Math.PI
        If Not (roundBetween(newA, roundAdd(newA, max, round), oldA, round) OrElse roundBetween(newA, roundSub(newA, max, round), oldA, round)) Then
            If roundDirection(newA, oldA, round) = 1 Then
                newA = roundSub(oldA, max, round)
            Else
                newA = roundAdd(oldA, max, round)
            End If
        End If
        Return newA
    End Function

    Sub New(mSpeed As Double, c As Color, loc As PointF, s As SizeF, d As Integer, t As Creep, retarget As Boolean, tower As Type)
        MyBase.New(New PointF(0, 0), c, loc, s, d, True, tower)
        target = t
        moveSpeed = mSpeed
        Dim a = MathHelper.getAngle(New Vector2(m_Location), New Vector2(target.getLocation))
        Dim v = MathHelper.getPointOnCircle(New Vector2(), a, moveSpeed)
        moveVec = v.toPointF
        Me.retarget = retarget
    End Sub

    ''' <summary>
    ''' Führt eine Addition von val und addVal durch. Falls das Ergebnis größer als max ist, beginnt die Methode wieder bei 0.
    ''' </summary>
    ''' <param name="val"></param>
    ''' <param name="addVal"></param>
    ''' <param name="max"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Function roundAdd(val As Double, addVal As Double, max As Double) As Double
        Return (val + addVal) Mod max
    End Function

    ''' <summary>
    ''' Führt eine Subtraktion von val und subVal durch. Falls das Ergebnis kleiner als 0 ist, beginnt die Methode wieder bei max.
    ''' </summary>
    ''' <param name="val"></param>
    ''' <param name="subVal"></param>
    ''' <param name="max"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Shared Function roundSub(val As Double, subVal As Double, max As Double) As Double
        Dim retVal As Double = val - subVal Mod max
        If retVal >= 0 Then Return retVal
        Return max + retVal
    End Function

    Public Shared Function roundBetween(val1 As Double, val2 As Double, testVal As Double, max As Double) As Boolean
        Dim dist As Double = roundDistance(val1, val2, max)
        Return roundDistance(val1, testVal, max) <= dist AndAlso roundDistance(val2, testVal, max) <= dist
    End Function

    Public Shared Function roundDistance(val1 As Double, val2 As Double, max As Double) As Double
        Dim retVal As Double = roundSub(val1, val2, max)
        If retVal > max / 2 Then Return max - retVal
        Return retVal
    End Function

    Public Shared Function roundDirection(val1 As Double, val2 As Double, max As Double) As Integer
        If val1 = val2 OrElse val1 = roundAdd(val2, max / 2, max) Then Return 0
        Dim dist As Double = roundSub(val1, val2, max)
        If dist < max / 2 Then Return -1
        Return 1
    End Function
End Class

Friend Class BombProjectile : Inherits Projectile
    Private steps As Integer
    Private maxSteps As Integer
    Private radius As Double

    Sub New(v As PointF, c As Color, loc As PointF, s As SizeF, d As Integer, maxFlight As Integer, r As Double, tower As Type)
        MyBase.New(v, c, loc, s, d, False, tower)
        maxSteps = maxFlight
        radius = r
    End Sub

    Friend Overrides Sub doStep()
        MyBase.doStep()
        steps += 1
        If steps = maxSteps Then
            Dim ents = world.getEntities
            For Each ent In ents
                If TypeOf ent Is Creep Then
                    ent.InUse = True
                    If MathHelper.getDistance(New Helpers.Mathematics.Vector2(ent.getLocation), New Helpers.Mathematics.Vector2(getLocation)) <= radius Then DirectCast(ent, Creep).takeDamage(damage, towerType)
                    ent.InUse = False
                End If
            Next
            Dim ex As New Explosion(m_Location, {color}, color, 20 + damage, 0.0F, False, False)
            world.spawnEntity(ex)
            world.despawnEntity(Me)
        End If
    End Sub
End Class