﻿Option Strict On
Option Infer On

Imports SlimDX.Direct2D
Imports SlimDX
Imports Helpers.Mathematics
Imports System.Reflection

Friend MustInherit Class WaySegment
    Private m_P1 As PointF
    Private m_P2 As PointF
    Protected way As Way
    Protected world As World

    Friend ReadOnly Property P1 As PointF
        Get
            Return m_P1
        End Get
    End Property

    Friend ReadOnly Property P2 As PointF
        Get
            Return m_P2
        End Get
    End Property

    Friend ReadOnly Property length As Double
        Get
            Return Math.Sqrt(Math.Pow(m_P1.X - m_P2.X, 2) + Math.Pow(m_P1.Y - m_P2.Y, 2))
        End Get
    End Property

    Friend Shared Function GetAllWaySegments() As Type()
        Dim currentAssembly = Assembly.GetExecutingAssembly
        Dim currentTypes() = currentAssembly.GetTypes
        Dim waySegments As New List(Of Type)
        Dim towers = Tower.GetAllTowers
        For Each t In currentTypes
            If t.IsSubclassOf(GetType(WaySegment)) AndAlso Not t Is GetType(EndWaySegment) Then
                If t.IsGenericTypeDefinition Then
                    For Each tower In towers
                        If tower IsNot GetType(SlownessTower) Then waySegments.Add(t.MakeGenericType(tower))
                    Next
                Else
                    waySegments.Add(t)
                End If
            End If
        Next
        Return waySegments.ToArray
    End Function

    Friend Function GetPoints(speed As Double) As PointF()
        Return GetPoints(speed, P1, P2)
    End Function

    Friend Function GetPoints(speed As Double, first As PointF, second As PointF) As PointF()
        Dim steps = Convert.ToInt32(Math.Sqrt((second.X - first.X) ^ 2 + (second.Y - first.Y) ^ 2) / speed)
        If steps > 0 Then
            Dim p(steps) As PointF
            p(0) = first
            p(steps) = second
            Dim sX = (second.X - first.X) / steps
            Dim sY = (second.Y - first.Y) / steps
            For i = 1 To steps - 1
                p(i) = New PointF(first.X + i * sX, first.Y + i * sY)
            Next
            Return p
        End If
        Return Nothing
    End Function

    Friend Overridable Sub onCreepReachStart(ByVal c As Creep)

    End Sub

    Friend Overridable Sub onCreepReachEnd(ByVal c As Creep)

    End Sub

    Friend Overridable Sub Draw(e As D2DControlset.D2DDrawEventArgs, offset As PointF)

    End Sub

    Sub New(ByVal w As Way, ByVal wo As World, ByVal p1 As PointF, ByVal p2 As PointF)
        way = w
        world = wo
        m_P1 = p1
        m_P2 = p2
    End Sub
End Class

Friend Class DefaultWaySegment
    Inherits WaySegment

    Sub New(ByVal w As Way, ByVal wo As World, ByVal p1 As PointF, ByVal p2 As PointF)
        MyBase.New(w, wo, p1, p2)
    End Sub
End Class

Friend Class EndWaySegment
    Inherits WaySegment

    Friend Overrides Sub onCreepReachEnd(c As Creep)
        c.takeDamage(c.getLife, Nothing)
        world.reduceHealth()
    End Sub

    Sub New(ByVal w As Way, ByVal wo As World, ByVal p1 As PointF, ByVal p2 As PointF)
        MyBase.New(w, wo, p1, p2)
    End Sub
End Class

Friend Class ShieldWaySegment : Inherits WaySegment
    Private shineIndex As Integer

    Friend Overrides Sub onCreepReachStart(c As Creep)
        Dim index = Array.IndexOf(way.segments, Me)
        If Not (index > 0 AndAlso TypeOf way.segments(index - 1) Is ShieldWaySegment) Then c.AktivateShield()
    End Sub

    Friend Overrides Sub onCreepReachEnd(c As Creep)
        Dim index = Array.IndexOf(way.segments, Me)
        If Not (index < way.segments.Length - 1 AndAlso TypeOf way.segments(index + 1) Is ShieldWaySegment) Then c.DeaktivateShield()
    End Sub

    Sub New(w As Way, wo As World, p1 As PointF, p2 As PointF)
        MyBase.New(w, wo, p1, p2)
    End Sub

    Friend Overrides Sub Draw(e As D2DControlset.D2DDrawEventArgs, offset As PointF)
        Dim s1 As New GradientStop
        s1.Color = New Color4(Color.Transparent)
        s1.Position = 0
        Dim s2 As New GradientStop
        s2.Color = New Color4(Color.FromArgb(100, Color.SkyBlue))
        s2.Position = 0.5F
        Dim s3 As New GradientStop
        s3.Color = New Color4(Color.Transparent)
        s3.Position = 1
        Dim stops As New GradientStopCollection(e.Target, {s1, s2, s3})
        Dim props As New LinearGradientBrushProperties
        Dim shineWidth = CInt(MathHelper.getDistance(New Helpers.Mathematics.Vector2(Me.P1), New Helpers.Mathematics.Vector2(Me.P2)) * 0.5)
        Dim angle = MathHelper.getAngle(New Helpers.Mathematics.Vector2(Me.P1), New Helpers.Mathematics.Vector2(Me.P2))
        Dim p1 = MathHelper.getPointOnCircle(New Helpers.Mathematics.Vector2(Me.P1), angle + Math.PI + 1.5, shineWidth * 1.5).toPointF
        Dim p2 = MathHelper.getPointOnCircle(New Helpers.Mathematics.Vector2(Me.P2), angle + 1.5, shineWidth * 1.5).toPointF
        Dim points() = GetPoints(1, p1, p2)
        props.StartPoint = points(shineIndex).Offset(offset)
        props.EndPoint = points(shineIndex + shineWidth).Offset(offset)
        Dim shineBrush As New LinearGradientBrush(e.Target, stops, props)
        Dim sProps As New StrokeStyleProperties
        sProps.DashCap = CapStyle.Round
        sProps.DashStyle = DashStyle.Solid
        sProps.EndCap = CapStyle.Round
        sProps.StartCap = CapStyle.Round
        sProps.LineJoin = LineJoin.Round
        Dim style As New StrokeStyle(e.Factory, sProps)
        Dim path As New PathGeometry(e.Factory)
        Dim sink = path.Open
        sink.BeginFigure(Me.P1.Offset(offset), FigureBegin.Hollow)
        sink.AddLine(Me.P2.Offset(offset))
        sink.EndFigure(FigureEnd.Open)
        sink.Close()
        e.Target.DrawGeometry(path, shineBrush, 24, style)
        style.Dispose()
        sink.Dispose()
        path.Dispose()
        shineBrush.Dispose()
        stops.Dispose()
        shineIndex += 8
        If shineIndex >= points.Length - shineWidth Then shineIndex = 0
    End Sub
End Class

Friend Class SpeedWaySegment : Inherits WaySegment
    Private arrowIndex As Integer
    Private arrowPoints() As PointF
    Private angle As Double
    Private steps As Integer

    Friend Overrides Sub onCreepReachStart(c As Creep)
        c.ApllySpeedModifier(2)
    End Sub

    Friend Overrides Sub onCreepReachEnd(c As Creep)
        c.RemoveSpeedModifier()
    End Sub

    Sub New(w As Way, wo As World, p1 As PointF, p2 As PointF)
        MyBase.New(w, wo, p1, p2)
        arrowIndex = 9
        angle = MathHelper.getAngle(New Helpers.Mathematics.Vector2(p1), New Helpers.Mathematics.Vector2(p2))
        Dim p1x = MathHelper.getPointOnCircle(New Helpers.Mathematics.Vector2(Me.P1), angle + Math.PI, 200).toPointF
        Dim p2x = MathHelper.getPointOnCircle(New Helpers.Mathematics.Vector2(Me.P2), angle, 200).toPointF
        arrowPoints = GetPoints(20, p1x, p2x)
        angle += Math.PI
    End Sub

    Friend Overrides Sub Draw(e As D2DControlset.D2DDrawEventArgs, offset As PointF)
        For i = 0 To 9
            If arrowIndex - i <= 10 OrElse arrowIndex - i >= arrowPoints.Length - 10 Then Continue For
            Dim brushName = "arrowBrush" & i
            If Not ColorManager.BrushRegistered(brushName) Then ColorManager.RegisterBrush(brushName, New SolidColorBrush(e.Target, New Color4(1 - i * 0.1F, 0, 1, 1)))
            Dim aP1 = MathHelper.getPointOnCircle(New Helpers.Mathematics.Vector2(arrowPoints(arrowIndex - i)), angle - 1, 10).toPointF.Offset(offset)
            Dim aP2 = arrowPoints(arrowIndex - i).Offset(offset)
            Dim aP3 = MathHelper.getPointOnCircle(New Helpers.Mathematics.Vector2(arrowPoints(arrowIndex - i)), angle + 1, 10).toPointF.Offset(offset)
            Dim pg As New PathGeometry(e.Factory)
            Dim sink = pg.Open
            sink.BeginFigure(aP1, FigureBegin.Hollow)
            sink.AddLines({aP2, aP3})
            sink.EndFigure(FigureEnd.Open)
            sink.Close()
            e.Target.DrawGeometry(pg, ColorManager.GetBrush(brushName), 4)
            sink.Dispose()
            pg.Dispose()
        Next
        steps += 1
        If steps Mod 2 = 0 Then arrowIndex += 1
        If arrowIndex = arrowPoints.Length Then arrowIndex = 9
    End Sub
End Class

Friend Class SlowWaySegment : Inherits WaySegment
    Private dotIndex As Integer
    Private dotPoints() As PointF
    Private angle As Double
    Private steps As Integer

    Friend Overrides Sub onCreepReachStart(c As Creep)
        c.ApllySpeedModifier(0.5)
    End Sub

    Friend Overrides Sub onCreepReachEnd(c As Creep)
        c.RemoveSpeedModifier()
    End Sub

    Sub New(w As Way, wo As World, p1 As PointF, p2 As PointF)
        MyBase.New(w, wo, p1, p2)
        dotIndex = 0
        angle = MathHelper.getAngle(New Helpers.Mathematics.Vector2(p1), New Helpers.Mathematics.Vector2(p2))
        Dim p1x = MathHelper.getPointOnCircle(New Helpers.Mathematics.Vector2(Me.P1), angle + Math.PI, 200).toPointF
        Dim p2x = MathHelper.getPointOnCircle(New Helpers.Mathematics.Vector2(Me.P2), angle, 200).toPointF
        dotPoints = GetPoints(20, p1x, p2x)
    End Sub

    Friend Overrides Sub Draw(e As D2DControlset.D2DDrawEventArgs, offset As PointF)
        For i = 0 To 9
            If dotIndex + i <= 10 OrElse dotIndex + i >= dotPoints.Length - 10 Then Continue For
            Dim brushName = "dotBrush" & i
            If Not ColorManager.BrushRegistered(brushName) Then ColorManager.RegisterBrush(brushName, New SolidColorBrush(e.Target, New Color4(1 - i * 0.1F, 1, 1, 0)))
            Dim aP1 = MathHelper.getPointOnCircle(New Helpers.Mathematics.Vector2(dotPoints(dotIndex + i)), angle - 1, 10).toPointF.Offset(offset)
            Dim aP2 = dotPoints(dotIndex + i).Offset(offset)
            Dim aP3 = MathHelper.getPointOnCircle(New Helpers.Mathematics.Vector2(dotPoints(dotIndex + i)), angle + 1, 10).toPointF.Offset(offset)
            Dim pg As New PathGeometry(e.Factory)
            Dim sink = pg.Open
            sink.BeginFigure(aP1, FigureBegin.Hollow)
            sink.AddLines({aP2, aP3})
            sink.EndFigure(FigureEnd.Open)
            sink.Close()
            e.Target.DrawGeometry(pg, ColorManager.GetBrush(brushName), 4)
            sink.Dispose()
            pg.Dispose()
        Next
        steps += 1
        If steps Mod 2 = 0 Then dotIndex -= 1
        If dotIndex = -1 Then dotIndex = dotPoints.Length - 10
    End Sub
End Class

Friend Class AbsorbingWaySegment(Of T As Tower) : Inherits WaySegment
    Private points() As PointF
    Private towerColor As Color

    Friend Overrides Sub onCreepReachStart(c As Creep)
        c.BlockTower(GetType(T))
    End Sub

    Friend Overrides Sub onCreepReachEnd(c As Creep)
        c.UnblockTower()
    End Sub

    Sub New(w As Way, wo As World, p1 As PointF, p2 As PointF)
        MyBase.New(w, wo, p1, p2)
        points = GetPoints(10)
        towerColor = Color.FromArgb(10, DirectCast(GetType(T).GetMethod("GetBaseColor").Invoke(Nothing, {}), Color))
    End Sub

    Friend Overrides Sub Draw(e As D2DControlset.D2DDrawEventArgs, offset As PointF)
        For Each p In points
            Dim brushName = "absorbing" & world.name & Array.IndexOf(way.segments, Me) & p.ToString & towerColor.ToString
            If Not ColorManager.BrushRegistered(brushName) Then
                Dim s1 As New GradientStop
                s1.Position = 0
                s1.Color = New Color4(towerColor)
                Dim s2 As New GradientStop
                s2.Position = 1
                s2.Color = New Color4(Color.Transparent)
                Dim stops As New GradientStopCollection(e.Target, {s1, s2})
                Dim props As New RadialGradientBrushProperties
                props.CenterPoint = p.Offset(offset)
                props.HorizontalRadius = 40
                props.VerticalRadius = 40
                ColorManager.RegisterBrush(brushName, New RadialGradientBrush(e.Target, stops, props))
                stops.Dispose()
            End If
            Dim el As New Ellipse
            el.Center = p.Offset(offset)
            el.RadiusX = 40
            el.RadiusY = 40
            e.Target.FillEllipse(ColorManager.GetBrush(brushName), el)
        Next
    End Sub
End Class